/**
 * Ext GWT 3.0.0-beta3 - Ext for GWT
 * Copyright(c) 2007-2011, Sencha, Inc.
 * licensing@sencha.com
 *
 * http://sencha.com/license
 */
package com.sencha.gxt.chart.client.chart;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import com.google.gwt.core.client.Scheduler;
import com.google.gwt.core.client.Scheduler.ScheduledCommand;
import com.google.gwt.event.shared.HandlerRegistration;
import com.google.gwt.user.client.Event;
import com.sencha.gxt.chart.client.chart.axis.Axis;
import com.sencha.gxt.chart.client.chart.axis.CartesianAxis;
import com.sencha.gxt.chart.client.chart.axis.TimeAxis;
import com.sencha.gxt.chart.client.chart.series.ScatterSeries;
import com.sencha.gxt.chart.client.chart.series.Series;
import com.sencha.gxt.chart.client.draw.DrawComponent;
import com.sencha.gxt.chart.client.draw.sprite.TextSprite;
import com.sencha.gxt.core.client.util.PrecisePoint;
import com.sencha.gxt.core.client.util.PreciseRectangle;
import com.sencha.gxt.data.shared.ListStore;
import com.sencha.gxt.data.shared.event.StoreAddEvent;
import com.sencha.gxt.data.shared.event.StoreClearEvent;
import com.sencha.gxt.data.shared.event.StoreDataChangeEvent;
import com.sencha.gxt.data.shared.event.StoreFilterEvent;
import com.sencha.gxt.data.shared.event.StoreHandlers;
import com.sencha.gxt.data.shared.event.StoreRecordChangeEvent;
import com.sencha.gxt.data.shared.event.StoreRemoveEvent;
import com.sencha.gxt.data.shared.event.StoreSortEvent;
import com.sencha.gxt.data.shared.event.StoreUpdateEvent;
import com.sencha.gxt.fx.client.easing.Default;
import com.sencha.gxt.fx.client.easing.EasingFunction;
import com.sencha.gxt.widget.core.client.tips.ToolTip;

/**
 * The Chart package provides the capability to visualize data.
 * 
 * @param <M> the data type to be used by this Chart
 */
public class Chart<M> extends DrawComponent {

  /**
   * Enumeration used for items positioned on the chart.
   */
  public enum Position {
    /**
     * Positioned at the top of the chart.
     */
    TOP,
    /**
     * Positioned at the bottom of the chart.
     */
    BOTTOM,
    /**
     * Positioned at the left of the chart.
     */
    LEFT,
    /**
     * Positioned at the right of the chart.
     */
    RIGHT
  }

  private boolean resizing = false;
  private boolean shadow = false;
  private double defaultInsets = 20;
  private int currentWidth;
  private int currentHeight;
  private PreciseRectangle bbox;
  private PreciseRectangle zoom = new PreciseRectangle(0, 0, 1, 1);
  private ListStore<M> store;
  private ListStore<M> substore;
  private double[] maxGutter = {0, 0};
  private List<Series<M>> series = new ArrayList<Series<M>>();
  private Map<Position, Axis<M, ?>> axes = new HashMap<Position, Axis<M, ?>>();
  private Legend<M> legend;
  private int lastLegend = -1;
  private Series<M> lastSeries;
  private Axis<M, ?> lastAxis;
  private EasingFunction animationEasing = new Default();
  private int animationDuration = 500;
  private StoreHandlers<M> storeHandlers;
  private HandlerRegistration handlerRegistration;
  private boolean deferred = false;

  /**
   * Creates a chart with default width and height.
   */
  public Chart() {
    this(1, 1);
  }

  /**
   * Creates a chart with the given width and height.
   * 
   * @param width the width of the chart
   * @param height the height of the chart
   */
  public Chart(int width, int height) {
    super(width, height);
    setPixelSize(width, height);

    storeHandlers = new StoreHandlers<M>() {

      @Override
      public void onAdd(StoreAddEvent<M> event) {
        redrawChart();
      }

      @Override
      public void onClear(StoreClearEvent<M> event) {
        redrawChart();
      }

      @Override
      public void onDataChange(StoreDataChangeEvent<M> event) {
        redrawChart();
      }

      @Override
      public void onFilter(StoreFilterEvent<M> event) {
        redrawChart();
      }

      @Override
      public void onRecordChange(StoreRecordChangeEvent<M> event) {
        redrawChart();
      }

      @Override
      public void onRemove(StoreRemoveEvent<M> event) {
        redrawChart();
      }

      @Override
      public void onSort(StoreSortEvent<M> event) {
        redrawChart();
      }

      @Override
      public void onUpdate(StoreUpdateEvent<M> event) {
        redrawChart();
      }
    };
  }

  /**
   * Adds an axis to the chart. Only one axis for each {@link Position} can be
   * added.
   * 
   * @param axis the axis to be added
   */
  public void addAxis(Axis<M, ?> axis) {
    axes.put(axis.getPosition(), axis);
    axis.setChart(this);
  }

  /**
   * Adds a {@link Series} to the chart.
   * 
   * @param series the series to be added
   */
  public void addSeries(Series<M> series) {
    this.series.add(series);
    series.setChart(this);
  }

  /**
   * Binds the store to the chart so that any store events will automatically
   * redraw the chart.
   * 
   * @param store the store to be bound
   */
  public void bindStore(ListStore<M> store) {
    if (this.store != null) {
      handlerRegistration.removeHandler();
    }
    if (store != null) {
      handlerRegistration = store.addStoreHandlers(storeHandlers);
    }
    setStore(store);
  }

  /**
   * Calculates the maximum gutter of the chart's series.
   */
  public void calculateMaximumGutter() {
    maxGutter[0] = 0;
    maxGutter[1] = 0;
    for (Series<M> series : this.series) {
      double gutter[] = series.getGutters();
      maxGutter[0] = Math.max(maxGutter[0], gutter[0]);
      maxGutter[1] = Math.max(maxGutter[1], gutter[1]);
    }
  }

  /**
   * Returns the duration of animations on the chart.
   * 
   * @return the duration of animations on the chart
   */
  public int getAnimationDuration() {
    return animationDuration;
  }

  /**
   * Returns the {@link EasingFunction} used by animations on the chart.
   * 
   * @return the easing function used by animations on the chart
   */
  public EasingFunction getAnimationEasing() {
    return animationEasing;
  }

  /**
   * Returns the map of axes currently added to the chart.
   * 
   * @return the map of axes currently added to the chart
   */
  public Map<Position, Axis<M, ?>> getAxes() {
    return axes;
  }

  /**
   * Returns the {@link Axis} at the given {@link Position}.
   * 
   * @param position the position of the axis
   * @return the axis at the given position
   */
  public Axis<M, ?> getAxis(Position position) {
    return axes.get(position);
  }

  /**
   * Returns the bounding box of the chart.
   * 
   * @return the bounding box of the chart
   */
  public PreciseRectangle getBBox() {
    return bbox;
  }

  /**
   * Returns the store currently being used by the chart.
   * 
   * @return the store currently being used by the chart
   */
  public ListStore<M> getCurrentStore() {
    if (substore != null) {
      return substore;
    } else {
      return store;
    }
  }

  /**
   * Returns the initial insets of the chart.
   * 
   * @return the initial insets of the chart
   */
  public double getDefaultInsets() {
    return defaultInsets;
  }

  /**
   * Returns the legend of the chart.
   * 
   * @return the legend of the chart
   */
  public Legend<M> getLegend() {
    return legend;
  }

  /**
   * Returns maximum gutter of the series in the chart.
   * 
   * @return maximum gutter of the series in the chart
   */
  public double[] getMaxGutter() {
    return maxGutter;
  }

  /**
   * Returns all of the series attached to the chart.
   * 
   * @return all of the series attached to the chart
   */
  public List<Series<M>> getSeries() {
    return series;
  }

  /**
   * Returns the series attached to the chart at the given index.
   * 
   * @param index the index of the series
   * @return the series attached to the chart
   */
  public Series<M> getSeries(int index) {
    return series.get(index);
  }

  /**
   * Returns the data store bound to the chart.
   * 
   * @return the data store bound to the chart
   */
  public ListStore<M> getStore() {
    return store;
  }

  /**
   * Returns the zoom of the chart.
   * 
   * @return the zoom of the chart
   */
  public PreciseRectangle getZoom() {
    return zoom;
  }

  /**
   * Returns whether or not the chart has shadows.
   * 
   * @return true if shadows
   */
  public boolean hasShadows() {
    return shadow;
  }

  /**
   * Returns whether or not the chart is resizing.
   * 
   * @return true if resizing
   */
  public boolean isResizing() {
    return resizing;
  }

  /**
   * Method used when the chart is clicked.
   * 
   * @param event the mouse event
   */
  @Override
  public void onMouseDown(Event event) {
    PrecisePoint point = getEventXY(event);
    if (lastLegend != -1) {
      legend.onMouseDown(lastLegend, event);
    }
    if (lastSeries != null) {
      lastSeries.onMouseDown(point, event);
    }
    if (lastAxis != null) {
      lastAxis.onMouseDown(point, event);
    }
  }

  /**
   * Method used when the mouse moves over the chart.
   * 
   * @param event the mouse event
   */
  @Override
  public void onMouseMove(Event event) {
    PrecisePoint point = getEventXY(event);
    if (legend != null) {
      final int item;
      if (legend.getBBox().contains(point)) {
        item = legend.getItemFromPoint(point);
      } else {
        item = -1;
      }
      if (lastLegend != -1 && lastLegend != item) {
        if (legend.getToolTip() != null) {
          legend.getToolTip().hide();
        }
        legend.onMouseOut(lastLegend, event);
      }
      if (item != -1) {
        ToolTip tip = legend.getToolTip();
        LegendItem<M> sprite = legend.getItems().get(item);
        if (tip != null) {
          LegendToolTipConfig<M> config = legend.getToolTipConfig();
          if (config.getCustomLabelProvider() != null) {
            config.setBodyText(config.getCustomLabelProvider().getLabel(sprite));
          } else if (sprite.size() > 0) {
            config.setBodyText(((TextSprite) sprite.get(0)).getText());
          }
          tip.update(config);
          tip.showAt(event.getClientX() + 20, event.getClientY() + 20);
        }
        legend.onMouseMove(item, event);
      }
      lastLegend = item;
    }

    for (Axis<M, ?> axis : this.axes.values()) {
      int index = axis.onMouseMove(point, event);
      if (index != -1) {
        lastAxis = axis;
      }
    }

    for (int i = 0; i < series.size(); i++) {
      Series<M> s = series.get(i);
      s.calculateBBox(false);
      double selectionTolerance = 0;
      if (s instanceof ScatterSeries) {
        selectionTolerance = ((ScatterSeries<M>) s).getSelectionTolerance();
      }
      if (s.getBBox().contains(point, selectionTolerance)) {
        int index = s.onMouseMove(point, event);
        if (index != -1) {
          lastSeries = s;
        }
      } else if (s.highlightedState()) {
        s.onMouseOut(point, event);
      }
    }
  }

  /**
   * Method used when the mouse leaves the chart.
   * 
   * @param event the mouse event
   */
  @Override
  public void onMouseOut(Event event) {
    if (!getElement().getBounds().contains(event.getClientX(), event.getClientY())) {
      PrecisePoint point = getEventXY(event);
      for (int i = 0; i < series.size(); i++) {
        series.get(i).onMouseOut(point, event);
      }
    }
  }

  @Override
  public void onResize(int width, int height) {
    this.resizing = true;
    setPixelSize(width, height);
    redrawChart();
    super.onResize(width, height);
  }

  /**
   * Redraws all elements of the chart.
   */
  public void redrawChart() {
    if (!deferred) {
      deferred = true;
      Scheduler.get().scheduleDeferred(new ScheduledCommand() {
        @Override
        public void execute() {
          render();
        }
      });
    }
  }

  /**
   * Redraws all elements of the chart immediatley.
   */
  public void redrawChartForced() {
    render();
  }

  /**
   * Removes the {@link Axis} from the chart.
   * 
   * @param axis the axis to remove
   * @return the removed axis
   */
  public Axis<M, ?> removeAxis(Axis<M, ?> axis) {
    if (axis != null) {
      axis.clear();
      axis.setChart(null);
      return axes.remove(axis.getPosition());
    } else {
      return null;
    }
  }

  /**
   * Removes the {@link Axis} at the given {@link Position} from the chart.
   * 
   * @param position the position of the axis to remove
   * @return the removed axis
   */
  public Axis<M, ?> removeAxis(Position position) {
    return removeAxis(axes.get(position));
  }

  /**
   * Removes the {@link Legend} currently attached legend from the chart.
   */
  public void removeLegend() {
    if (legend != null) {
      legend.clear();
      legend.setChart(null);
      legend = null;
    }
  }

  /**
   * Removes the given {@link Series} from the chart.
   * 
   * @param index the index of the series to be removed
   * @return true if the series is removed
   */
  public boolean removeSeries(int index) {
    return removeSeries(getSeries(index));
  }

  /**
   * Removes the given {@link Series} from the chart.
   * 
   * @param series the series to be removed
   * @return true if the series is removed
   */
  public boolean removeSeries(Series<M> series) {
    series.clear();
    series.setChart(null);
    return this.series.remove(series);
  }

  /**
   * Sets the duration of animations on the chart.
   * 
   * @param animationDuration the duration of animations on the chart
   */
  public void setAnimationDuration(int animationDuration) {
    this.animationDuration = animationDuration;
  }

  /**
   * Sets the {@link EasingFunction} used by animations on the chart.
   * 
   * @param animationEasing the easing function used by animations on the chart
   */
  public void setAnimationEasing(EasingFunction animationEasing) {
    this.animationEasing = animationEasing;
  }

  /**
   * Sets whether or not the chart has shadows.
   * 
   * @param shadow true if shadows
   */
  public void setChartShadow(boolean shadow) {
    this.shadow = shadow;
  }

  /**
   * Sets the initial insets of the chart.
   * 
   * @param defaultInsets the initial insets of the chart
   */
  public void setDefaultInsets(double defaultInsets) {
    this.defaultInsets = defaultInsets;
  }

  /**
   * Sets the legend of the chart.
   * 
   * @param legend the legend of the chart
   */
  public void setLegend(Legend<M> legend) {
    legend.setChart(this);
    this.legend = legend;
  }

  @Override
  public void setPixelSize(int width, int height) {
    super.setPixelSize(width, height);
    if (currentWidth != width && currentHeight != height) {
      bbox = new PreciseRectangle(0, 0, width, height);
    }
    currentWidth = width;
    currentHeight = height;
  }

  /**
   * Sets the data store bound to the chart.
   * 
   * @param store the data store
   */
  public void setStore(ListStore<M> store) {
    this.store = store;
  }

  /**
   * Sets a temporary substore for the chart. Used by {@link TimeAxis}.
   * 
   * @param substore substore for the chart
   */
  public void setSubstore(ListStore<M> substore) {
    this.substore = substore;
  }

  /**
   * Sets the zoom of the chart.
   * 
   * @param zoom the zoom of the chart
   */
  public void setZoom(PreciseRectangle zoom) {
    this.zoom = zoom;
  }

  /**
   * Adjust the dimensions and positions of each axis and the chart body area
   * after accounting for the space taken up on each side by the axes and
   * legend.
   */
  private void alignAxes() {
    Position[] edges = Position.values();
    Map<Position, Double> insets = new HashMap<Position, Double>();
    // Find the space needed by axes and legend as a positive inset from each
    // edge
    for (int i = 0; i < edges.length; i++) {
      boolean isVertical = (edges[i] == Position.LEFT || edges[i] == Position.RIGHT);
      Axis<M, ?> axis = getAxis(edges[i]);
      double inset = defaultInsets;

      // Add legend size if it's on this edge
      if (legend != null) {
        if (legend.getPosition() == edges[i]) {
          PreciseRectangle bbox = legend.getBBox();
          if (isVertical) {
            inset += bbox.getWidth() + inset;
          } else {
            inset += bbox.getHeight() + inset;
          }
        }
      }

      // Add axis size if there's one on this edge
      if (axis != null) {
        PreciseRectangle bbox = axis.getBBox();
        if (isVertical) {
          inset += bbox.getWidth();
        } else {
          inset += bbox.getHeight();
        }
      }

      insets.put(edges[i], inset);
    }
    // Build the chart bbox based on the collected inset values
    bbox.setX(insets.get(Position.LEFT));
    bbox.setY(insets.get(Position.TOP));
    bbox.setWidth(currentWidth - insets.get(Position.LEFT) - insets.get(Position.RIGHT));
    bbox.setHeight(currentHeight - insets.get(Position.TOP) - insets.get(Position.BOTTOM));

    // Go back through each axis and set its length and position based on the
    // corresponding edge of the chartBBox
    for (Axis<M, ?> ax : this.axes.values()) {
      if (ax instanceof CartesianAxis) {
        CartesianAxis<M, ?> axis = (CartesianAxis<M, ?>) ax;
        Position pos = axis.getPosition();
        boolean isVertical = (pos == Position.LEFT || pos == Position.RIGHT);
        axis.setX((pos == Position.RIGHT) ? (bbox.getX() + bbox.getWidth()) : (bbox.getX()));
        axis.setY((pos == Position.TOP) ? (bbox.getY()) : (bbox.getY() + bbox.getHeight()));
        axis.setDepth(isVertical ? bbox.getWidth() : bbox.getHeight());
        axis.setLength(isVertical ? bbox.getHeight() : bbox.getWidth());
      }
    }
  }

  /**
   * Creates and initializes the default axis.
   * 
   * @param axis the axis to be initialized
   */
  private void initializeAxis(Axis<M, ?> ax) {
    if (ax instanceof CartesianAxis) {
      CartesianAxis<M, ?> axis = (CartesianAxis<M, ?>) ax;
      double w = bbox.getWidth();
      double h = bbox.getHeight();
      double x = bbox.getX();
      double y = bbox.getY();
      switch (axis.getPosition()) {
        case TOP:
          axis.setLength(w);
          axis.setDepth(h);
          axis.setX(x);
          axis.setY(y);
          break;
        case BOTTOM:
          axis.setLength(w);
          axis.setDepth(h);
          axis.setX(x);
          axis.setY(h);
          break;
        case LEFT:
          axis.setLength(h);
          axis.setDepth(w);
          axis.setX(x);
          axis.setY(h);
          break;
        case RIGHT:
          axis.setLength(h);
          axis.setDepth(w);
          axis.setX(w);
          axis.setY(h);
          break;
      }
    }
    ax.drawAxis(true);
  }

  private void render() {
    // Instantiate Series and Axes
    if (surface.getSurfaceElement() == null) {
      surface.draw();
    }
    for (Axis<M, ?> axis : this.axes.values()) {
      initializeAxis(axis);
    }

    for (Axis<M, ?> axis : this.axes.values()) {
      axis.drawAxis(false);
    }

    // Create legend if not already created
    if (legend != null) {
      legend.create();
    }

    // Place axes properly, including influence from each other
    this.alignAxes();

    // Reposition legend based on new axis alignment
    if (legend != null) {
      legend.updatePosition();
    }

    // Find the max gutter
    this.calculateMaximumGutter();

    // Draw axes and series
    for (Axis<M, ?> axis : this.axes.values()) {
      axis.drawAxis(false);
    }
    for (Series<M> series : this.series) {
      series.drawSeries();
    }
    if (!resizing) {
      super.redrawSurface();
    }

    deferred = false;
    resizing = false;
  }

}
